
/*****************************************************************************/
/*                                                                           */
/*  Practica de Arquitectures Paral.leles                                    */
/*                                                                           */
/*  Execucio paral.lela del codi basada en els estandards de OpenMP          */
/*  Versio amb algorisme millorat, optimitzada per a 2 threads               */
/*                                                                           */
/*  Realitzada per: Francisco Jose Aguilar Celdran i Sergio Blanco Cuaresma  */
/*                                                                           */
/*****************************************************************************/ 

#include<stdio.h>
#include<assert.h>

#define N 70000
#define NUM_TOTAL_THREADS 2

int valors[N];
int resultat[N];

main()
{
   int i,j,min_index,minim;
   double sum=0;
   FILE *ent, *sal;
   
   int thread_actual = -1;
   int limit_vector;
   int punt_mitg_vector;
   int index[2];
   int N_senar = 1;
   
   // Algorisme especialitzat en 2 threads
   omp_set_num_threads(2);
   
   for(i=0;i<N;i++) valors[i]= rand();
   
   ent=fopen("sort2.ent","w");
   assert(ent!=NULL);
   
   for(i=0;i<N;i++) fprintf(ent,"%d\n",valors[i]);
   close(ent);
   
   // Comprovaci per saber si podrem fer divisions tipus (N-1)/2
   if (N%2 == 0) {
      N_senar = 0; // N-1 no es divisible por 2
   }
   
   // Si el nombre d'iteracions es senar, l'ltim thread es qui s'emporta la
   // la iteraci restant
   #pragma omp parallel for schedule(static) private(minim,min_index,j,thread_actual,limit_vector)
   for (i=0;i<N-1;i++) 
   {
      minim = valors[i+1];
      min_index = i+1;
      
      thread_actual = omp_get_thread_num();  // privat per cada thread
      
      if (thread_actual + 1 == NUM_TOTAL_THREADS) {
         // L'ultim thread ha d'anar fins al final del vector
         limit_vector = N;
      } else {
         // Aquest no es l'ultim thread generat
         // Per calcular el limit del subvector tenim en compte les consideracions anteriors
         if (N_senar) {
            limit_vector = ((N-1)/NUM_TOTAL_THREADS) * (thread_actual+1);
         } else {
            limit_vector = ((N)/NUM_TOTAL_THREADS) * (thread_actual+1);
         }
      }

      if ((i < (limit_vector-1)) || (thread_actual + 1 == NUM_TOTAL_THREADS)) {
         for (j=i+2;j<limit_vector;j++){
            if (valors[j] < minim) 
            {
               minim = valors[j];
               min_index = j;
            }
         }
         if (valors[i] > minim) 
         {
            valors[min_index] = valors[i];
            valors[i] = minim;
         }
      }
   }
   
   // Fusionem els 2 subvectors en 1 vector nou anomenat resultat[]
   
   // Inicialitzacions
   i = 0;
   index[0] = 0;
   if ((N-1) % 2 == 0) {                      // N-1 es divisible per 2
      index[1] = ((N-1)/NUM_TOTAL_THREADS);
   } else {
      index[1] = ((N)/NUM_TOTAL_THREADS);
   }
   punt_mitg_vector = index[1];
   
   // Fusi
   while (i < N) {
      // Si no s'acabat de recorre els 2 subvectors
      if ((index[0] < punt_mitg_vector) && (index[1] < N)) {
         while ((index[0] < punt_mitg_vector) && (valors[index[0]] <= valors[index[1]])) {
            resultat[i] = valors[index[0]];
            i++;
            index[0]++;
         }
         while ((index[1] < N) && (valors[index[0]] >= valors[index[1]])) {
            resultat[i] = valors[index[1]];
            i++;
            index[1]++;
            
         }
      } else if (index[0] < punt_mitg_vector) { // Si s'ha acabat de recorre el segon subvector
         // Copiar tot el subvector que falta
         while (i<N) {
            resultat[i] = valors[index[0]];
            i++;
            index[0]++;
         }
      } else if (index[1] < N) { // Si s'ha acabat de recorre el primer subvector
         // Copiar tot el subvector que falta
         while (i<N) {
            resultat[i] = valors[index[1]];
            i++;
            index[1]++;
         }
      }
      
   }
   
   sal=fopen("para.sal","w");
   assert(sal!=NULL);
  
   // Amb la clausula reduction(+:sum) es crea una variable 'sum' per
   // a cada thread, i posteriorment es realitza la suma de totes elles
   #pragma omp parallel for reduction(+:sum)
   for (i=0;i<N;i++) {
      if (i%100) sum += resultat[i];
      //fprintf(sal,"%i\n",resultat[i]);
   }
  
   fprintf(sal,"Suma: %f\n",sum);
   //for(i=0;i<N;i++) fprintf(sal,"%d\n",valors[i]);
   close(sal);

}


